/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This software is Copyright  2011 Yuri Gonzaga Gonalves da Costa <yuriggc at gmail>, and it is hereby released to the general public under the following terms:*
 * Redistribution and use in source and binary forms, with or without modification, are permitted.                                                                 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
parameter DATA_WIDTH=32;
parameter ADDR_WIDTH=10;
module ExpandKey(		
	input wire reset,
	input wire clk,

	/*Input Handshake*/	
	input  wire  VALID_E,
	output logic READY_E,

	/*Output Handshake*/
	output logic VALID_S,
	input  wire  READY_S, 
	
	input wire unsigned [31:0]data_in,
	input wire unsigned [2:0]code,
	
	output  logic unsigned[191:0]data_out
	
);
	logic unsigned  [31:0]L;
	logic unsigned  [31:0]R;
	logic unsigned  [31:0]F_r;
	logic unsigned  [7:0]key_r[71:0];
	logic unsigned  [7:0]key_r2[71:0];
	logic unsigned 	[6:0]n_key_r;
	logic unsigned 	[6:0]n_key_r2;
    logic unsigned	[6:0]n_ciclic_key;
	logic unsigned  [127:0]salt_r;
	logic unsigned  [127:0]salt_r2;
	logic unsigned  [31:0]cost_r;
	logic ctrl; 
	
	logic unsigned  [63:0]pApB;	
		
	logic unsigned  [31:0]parray[17:0];
	logic unsigned	[3:0] count_round;
	logic unsigned	[4:0] count_XOR_PARRAY;
	logic unsigned  [9:0] count_XOR_SBOX;
	
	logic S1S2;
	
	
	/* State machines states */
	
	enum logic[4:0] {
		IN_HANDSHAKE 					,
		FEISTEL							,
		F1 		  						,
		PREF							,
		XOR_P17_P18                		,
		OUT_HANDSHAKE				   	,
		NOP1						   	,
		XOR_XR_SWAP						,
		SWAPR							,
		NOP2						   	,
		INIT_XOR						,
		CONTROL_EXPAND_KEY				,
		INIT_PARRAY,
		INIT_SBOXES,
		CYCLIC_KEY,
		F2,
		F0
	} state = IN_HANDSHAKE;
	
	
	enum logic[3:0] {
		XOR_PARRAY_1,
		XOR_PARRAY_2,
		XOR_SBOX_1,
		XOR_SBOX_2,
		COST_1,
		COST_2,
		COST_3
	} controlstate;

	/*RAM*/
	logic [(DATA_WIDTH-1):0] data_a, data_b;
	logic [(ADDR_WIDTH-1):0] addr_a, addr_b;
	logic we_a, we_b, clk_a, clk_b;
	wire [(DATA_WIDTH-1):0] q_a, q_b;
	RAM ramb(.*);
	
	/*ROM*/
	logic [(ADDR_WIDTH-1):0] addr_ar, addr_br; 
	wire [(DATA_WIDTH-1):0] q_ar, q_br;
	ROM romb(addr_ar,addr_br,clk,q_ar,q_br);
	
	assign clk_a = clk;
	assign clk_b = clk;
	
	logic accessingMEM;
	always_ff@(posedge clk) begin
		if(reset==1'b1) begin
			if(accessingMEM == 1'b0)begin
				if(state == IN_HANDSHAKE) begin//ready to read input data
					if(VALID_E ==1'b1)begin//valid data to be read
						if(code == 3'd0)begin
							cost_r <= 32'd1<<data_in;//cost
						end 
						else if(code == 3'd1)begin
							salt_r[127:96] <= data_in;
							salt_r2[127:96]	<= data_in;
						end
						else if(code == 3'd2)begin
							salt_r[95:64]   <= data_in;
							salt_r2[95:64]	<= data_in;
							n_key_r  <= 7'd0 ;
							n_key_r2 <= 7'd0;
						end
						else if(code == 3'd3)begin
							salt_r[63:32] <= data_in;
							salt_r2[63:32]	<= data_in;
						end
						else if(code == 3'd4)begin
							salt_r[31:0]   <= data_in;
							salt_r2[31:0]	<= data_in;
							n_key_r  <= 7'd0 ;
							n_key_r2 <= 7'd0;
						end
						else if(code == 3'd5)begin
							key_r[n_key_r]  <= data_in[7:0];
							key_r2[n_key_r] <= data_in[7:0];
							n_key_r <= n_key_r + 7'd1 ;
							n_key_r2 <= n_key_r2 + 7'd1;
						end 
						else if(code == 3'd6)begin
							key_r[n_key_r]  <= data_in[7:0];
							key_r2[n_key_r] <= data_in[7:0];
							n_key_r <= n_key_r + 7'd1 ;
							n_key_r2 <= n_key_r2 + 7'd1;
							
							ctrl <= 1'b0;
							S1S2 <= 1'b0;
							READY_E <= 1'b0;
							
							addr_ar <= 10'd0;
							addr_br <= 10'd1;
							
							we_a <= 1'b0;
							we_b <= 1'b0;
							
							count_round <= 4'd0;
							state <= INIT_PARRAY; 
						end
					end
				end if(state == INIT_PARRAY)begin
				/*parray initialization*/
					parray[0] <= 32'h243f6a88;
					parray[1] <= 32'h85a308d3;
					parray[2] <= 32'h13198a2e;
					parray[3] <= 32'h3707344;
					parray[4] <= 32'ha4093822;
					parray[5] <= 32'h299f31d0;
					parray[6] <= 32'h82efa98;
					parray[7] <= 32'hec4e6c89;
					parray[8] <= 32'h452821e6;
					parray[9] <= 32'h38d01377;
					parray[10] <= 32'hbe5466cf;
					parray[11] <= 32'h34e90c6c;
					parray[12] <= 32'hc0ac29b7;
					parray[13] <= 32'hc97c50dd;
					parray[14] <= 32'h3f84d5b5;
					parray[15] <= 32'hb5470917;
					parray[16] <= 32'h9216d5d9;
					parray[17] <= 32'h8979fb1b;
					
					state <= INIT_SBOXES;
				end
				else if(state == INIT_SBOXES ) begin/*Supondo que o primeiro e segundo dados ja foram  lidos da rom*/
					if( addr_br == 10'd1023) begin
						/*Endereco a ser escrito na ram*/
						addr_a <= addr_ar;
						addr_b <= addr_br;
						
						/*dados a serem escritosas ra,*/
						data_a <= q_ar;
						data_b <= q_br;
						
						/*Autoriza escrita na ram*/
						we_a   <= 1'b1;
						we_b   <= 1'b1;
						
						/*J foram lidos todos valores da ROM*/
						addr_ar <= addr_ar + 10'd0;
						addr_br <= addr_br  + 10'd0;
						
						/*No mesmo ciclo eh feita a leitura da rom e escrita na ram*/
						accessingMEM <= 1'b1;
						/*Apos ultima escrita na ram, ela estara inicializada*/
						state <= NOP1;
					end
					else begin
						/*Endereco a ser escrito na ram*/
						addr_a <= addr_ar;
						addr_b <= addr_br;
						
						/*dados a serem escritosas ra,*/
						data_a <= q_ar;
						data_b <= q_br;
						
						/*Autoriza escrita na ram*/
						we_a   <= 1'b1;
						we_b   <= 1'b1;
						
						/*Novos valores a serem lidos da ROM*/
						addr_ar <= addr_ar + 10'd2;
						addr_br <=addr_br  + 10'd2;
						
						/*No mesmo ciclo eh feita a leitura da rom e escrita na ram */
						
						accessingMEM <= 1'b1;
					end
				end
				else if(state == NOP1) begin
					n_ciclic_key <= n_key_r2;
					n_key_r <= 7'd0;
					state <= CYCLIC_KEY;
				end 
				else if( state == CYCLIC_KEY)begin
						if( n_ciclic_key < 7'd72)begin
							key_r[n_ciclic_key]  <=key_r[n_key_r];
							key_r2[n_ciclic_key] <=key_r[n_key_r];
							n_ciclic_key <= n_ciclic_key + 7'd1;
							if(n_key_r == (n_key_r2-7'd1))begin
								n_key_r <= 7'd0;
							end
							else begin
								n_key_r <= n_key_r + 7'd1;
							end
						end 					
						else begin
							n_key_r <= 7'd0;
							n_key_r2 <= 7'd72;
							count_XOR_PARRAY <= 5'd0;
							state <= INIT_XOR;
						end
					end
				else if(state == INIT_XOR) begin
					if(count_XOR_PARRAY < 5'd18)begin
						parray[count_XOR_PARRAY][31:24] <= parray[count_XOR_PARRAY][31:24]^ key_r[n_key_r];
						parray[count_XOR_PARRAY][23:16] <= parray[count_XOR_PARRAY][23:16]^ key_r[n_key_r+(7'd1)];
						parray[count_XOR_PARRAY][15:8]  <= parray[count_XOR_PARRAY][15:8] ^ key_r[n_key_r+(7'd2)];
						parray[count_XOR_PARRAY][ 7:0]  <= parray[count_XOR_PARRAY][ 7:0] ^ key_r[n_key_r+(7'd3)];
						count_XOR_PARRAY <= count_XOR_PARRAY + 5'd1;
						n_key_r <= n_key_r + 7'd4;
					end
					else begin
						n_key_r <= n_key_r2;
						count_XOR_PARRAY <= 5'd0;
						pApB <= 64'h0000000000000000;
						controlstate <= XOR_PARRAY_1;
						state <= CONTROL_EXPAND_KEY;
					end
				end
				else if(state == CONTROL_EXPAND_KEY) begin
					if(controlstate == XOR_PARRAY_1) begin
						if( S1S2 == 1'b0 )begin
							L <= salt_r[127:96]^pApB[63:32];//xL = xL ^ pArray[count_pArray];
							R <= salt_r[95:64]^pApB[31:0];
							S1S2 <= 1'b1;
						end
						else begin
							L 		<= salt_r[63:32] ^pApB[63:32];//xL = xL ^ pArray[count_pArray];
							R 		<= salt_r[31:0]^pApB[31:0];
							S1S2 	<= 1'b0;
						end
						count_round <= 4'd0;
						controlstate <= XOR_PARRAY_2;
						state <= PREF;
					end
					else if(controlstate == XOR_PARRAY_2)begin
						parray[count_XOR_PARRAY] <= pApB[63:32];
						parray[count_XOR_PARRAY+5'd1] <= pApB[31:0];
						if(count_XOR_PARRAY < 5'd16) begin
							count_XOR_PARRAY <= count_XOR_PARRAY+5'd2;
							controlstate <= XOR_PARRAY_1;
						end
						else begin
							count_XOR_SBOX <= 10'd0;
							controlstate <= XOR_SBOX_1;
						end
					end
					else if( controlstate == XOR_SBOX_1)begin
						we_a   <= 1'b0;
						we_b   <= 1'b0;
						if( S1S2 == 1'b0 )begin
							L 	<= salt_r[127:96]^pApB[63:32];//xL = xL ^ pArray[count_pArray];
							R 		<= salt_r[95:64]^pApB[31:0];
							S1S2 	<= 1'b1;
						end
						else begin
							L 	 <= salt_r[63:32]^pApB[63:32];//xL = xL ^ pArray[count_pArray];
							R 	 <= salt_r[31:0]^pApB[31:0];
							S1S2 <= 1'b0;
						end
						count_round <= 4'd0;
						controlstate <= XOR_SBOX_2;
						state <= PREF;
					end
					else 
					if(controlstate == XOR_SBOX_2)begin
						/*Atualizacao de sboxes*/
						addr_a <= count_XOR_SBOX;
						addr_b <= count_XOR_SBOX+10'd1;
						data_a <= pApB[63:32];
						data_b <= pApB[31:0];
						we_a   <= 1'b1;
						we_b   <= 1'b1;
						//accessingMEM = 1'b1;
						if(count_XOR_SBOX == 10'd1022) begin
							salt_r <=128'h00000000000000000000000000000000;
							controlstate <= COST_1;
						end
						else begin
							count_XOR_SBOX <= count_XOR_SBOX+10'd2;
							controlstate <= XOR_SBOX_1;
						end
					end else  
					if(controlstate ==  COST_1)begin
						we_a   <= 1'b0;
						we_b   <= 1'b0;
						if(cost_r==32'd0)begin
							VALID_S <= 1'b1;
							state <= OUT_HANDSHAKE;
						end else if(ctrl==1'b0)begin
							key_r <= key_r2;
							n_key_r <= 7'd0;
							ctrl <= 1'b1;
							S1S2 <= 1'b0;
							
							count_XOR_PARRAY <= 5'd0;
							pApB <= 64'h0000000000000000;
							
							controlstate <=  XOR_PARRAY_1;
							state <= INIT_XOR;
						end else begin
							key_r[7'd0] <= salt_r2[127:120];
							key_r[7'd1] <= salt_r2[119:112];
							key_r[7'd2] <= salt_r2[111:104];
							key_r[7'd3] <= salt_r2[103:96];
							key_r[7'd4] <= salt_r2[95:88];
							key_r[7'd5] <= salt_r2[87:80];
							key_r[7'd6] <= salt_r2[79:72];
							key_r[7'd7] <= salt_r2[71:64];
							key_r[7'd8] <= salt_r2[63:56];
							key_r[7'd9] <= salt_r2[55:48];
							key_r[7'd10] <= salt_r2[47:40];
							key_r[7'd11] <= salt_r2[39:32];
							key_r[7'd12] <= salt_r2[31:24];
							key_r[7'd13] <= salt_r2[23:16];
							key_r[7'd14] <= salt_r2[15:8];
							key_r[7'd15] <= salt_r2[7:0];
							
							key_r[7'd16] <= salt_r2[127:120];
							key_r[7'd17] <= salt_r2[119:112];
							key_r[7'd18] <= salt_r2[111:104];
							key_r[7'd19] <= salt_r2[103:96];
							key_r[7'd20] <= salt_r2[95:88];
							key_r[7'd21] <= salt_r2[87:80];
							key_r[7'd22] <= salt_r2[79:72];
							key_r[7'd23] <= salt_r2[71:64];
							key_r[7'd24] <= salt_r2[63:56];
							key_r[7'd25] <= salt_r2[55:48];
							key_r[7'd26] <= salt_r2[47:40];
							key_r[7'd27] <= salt_r2[39:32];
							key_r[7'd28] <= salt_r2[31:24];
							key_r[7'd29] <= salt_r2[23:16];
							key_r[7'd30] <= salt_r2[15:8];
							key_r[7'd31] <= salt_r2[7:0];
							
							key_r[7'd32] <= salt_r2[127:120];
							key_r[7'd33] <= salt_r2[119:112];
							key_r[7'd34] <= salt_r2[111:104];
							key_r[7'd35] <= salt_r2[103:96];
							key_r[7'd36] <= salt_r2[95:88];
							key_r[7'd37] <= salt_r2[87:80];
							key_r[7'd38] <= salt_r2[79:72];
							key_r[7'd39] <= salt_r2[71:64];
							key_r[7'd40] <= salt_r2[63:56];
							key_r[7'd41] <= salt_r2[55:48];
							key_r[7'd42] <= salt_r2[47:40];
							key_r[7'd43] <= salt_r2[39:32];
							key_r[7'd44] <= salt_r2[31:24];
							key_r[7'd45] <= salt_r2[23:16];
							key_r[7'd46] <= salt_r2[15:8];
							key_r[7'd47] <= salt_r2[7:0];
							
							
							key_r[7'd48] <= salt_r2[127:120];
							key_r[7'd49] <= salt_r2[119:112];
							key_r[7'd50] <= salt_r2[111:104];
							key_r[7'd51] <= salt_r2[103:96];
							key_r[7'd52] <= salt_r2[95:88];
							key_r[7'd53] <= salt_r2[87:80];
							key_r[7'd54] <= salt_r2[79:72];
							key_r[7'd55] <= salt_r2[71:64];
							key_r[7'd56] <= salt_r2[63:56];
							key_r[7'd57] <= salt_r2[55:48];
							key_r[7'd58] <= salt_r2[47:40];
							key_r[7'd59] <= salt_r2[39:32];
							key_r[7'd60] <= salt_r2[31:24];
							key_r[7'd61] <= salt_r2[23:16];
							key_r[7'd62] <= salt_r2[15:8];
							key_r[7'd63] <= salt_r2[7:0];
							
							key_r[7'd64] <= salt_r2[127:120];
							key_r[7'd65] <= salt_r2[119:112];
							key_r[7'd66] <= salt_r2[111:104];
							key_r[7'd67] <= salt_r2[103:96];
							key_r[7'd68] <= salt_r2[95:88];
							key_r[7'd69] <= salt_r2[87:80];
							key_r[7'd70] <= salt_r2[79:72];
							key_r[7'd71] <= salt_r2[71:64];
							
							n_key_r <= 7'd0;
							ctrl <= 1'b0;
							S1S2 <= 1'b0;
							cost_r <= cost_r - 32'd1;
							
							count_XOR_PARRAY <= 5'd0;
							pApB <= 64'h0000000000000000;
							controlstate <=  XOR_PARRAY_1;
							state <= INIT_XOR;
						end
					end
				end
				else if(state == PREF) begin 
					L <= L^parray[count_round];//xL = xL ^ pArray[count_round];
					state <= F0;
				end if(state == F0)begin
					/*Ler duas posicoes da sboxes*/
					addr_a <= L[31:24];
					addr_b <= 10'd256+L[23:16];
					
					accessingMEM <= 1'b1;
					state <= F1;
				end
				else if (state == F1)begin
					if( (q_a+q_b) < 33'd4294967296)begin
						//F_r <= (sboxes[L[31:24]]+sboxes[10'd256+L[23:16]]) ^ sboxes[10'd512+L[15:8]];
						F_r <= (q_a+q_b);
					end 
					else begin
						F_r <= (q_a+q_b - 33'd4294967296);
					end
					/*Ler mais duas posicoes da sboxes*/
					addr_a <= 10'd512+L[15:8];
					addr_b <= 10'd768+L[7:0];
					
					accessingMEM <= 1'b1;
					
					state <= F2;
				end
				else if (state == F2)begin
					if( ( (F_r^q_a) + q_b) < 33'd4294967296)begin
						F_r <= ((F_r^q_a) + q_b);
					end 
					else begin
						F_r <= ((F_r^q_a) + q_b) - 33'd4294967296;
					end
					state <= XOR_XR_SWAP;
				end
				else if(state == XOR_XR_SWAP)begin
					L <= R^F_r;
					R <= L;
					if(count_round<4'd15)begin
						count_round <= count_round+4'd1;
						state <=PREF;
					end
					else begin
						state <= XOR_P17_P18;
					end
				end
				else if(state == XOR_P17_P18) begin
					pApB[63:32] <=  R ^  parray[5'd17];
					pApB[31:0]  <=  L ^  parray[5'd16];
					state <= CONTROL_EXPAND_KEY;
				end
				else if(state == OUT_HANDSHAKE)begin
					if(READY_S==1'b1) begin
						READY_E <= 1'b1;
						VALID_S <= 1'b0;
						state <= IN_HANDSHAKE;
					end
				end
				//
				end 
			else begin
				/*leituras na memorira*/
				if(we_a) begin
					we_a <= 1'b0;
					we_b <= 1'b0;
				end
				accessingMEM <= 1'b0;
			end
		end	
		else begin
			/* Reset everything here */
			count_round <= 4'd0;
			state    <=   IN_HANDSHAKE;
			READY_E   <=  1'b1;
			VALID_S   <=  1'b0;
			accessingMEM <= 1'b0;
		end
	end
endmodule
